from __future__ import annotations

import typing as t
from dataclasses import dataclass

import clypi
from clypi import boxed, stack
from clypi._cli import type_util

if t.TYPE_CHECKING:
    from clypi.cli import Argument, SubCommand


@dataclass
class ProgramConfig:
    prog: str


def _ext(ls: list[str], s: str | list[str] | None) -> list[str] | str | None:
    if s is None:
        return ls
    if isinstance(s, str):
        ls.append(s)
    else:
        ls.extend(s)


def _pretty_traceback(err: BaseException) -> list[str]:
    # Get the traceback bottom up
    tb: list[BaseException] = [err]
    while tb[-1].__cause__ is not None:
        tb.append(tb[-1].__cause__)

    lines: list[str] = []
    for i, e in enumerate(reversed(tb)):
        icon = "  " * (i - 1) + " ↳ " if i != 0 else ""
        s = clypi.style(f"{icon}{str(e)}", fg="red")
        lines.append(s)
    return lines


@dataclass
class TermFormatter:
    prog: list[str]
    description: str | None
    epilog: str | None
    options: list[Argument]
    positionals: list[Argument]
    subcommands: list[SubCommand]
    exception: Exception | None

    def _format_option(self, option: Argument) -> tuple[str, ...]:
        usage = clypi.style(option.display_name, fg="blue", bold=True)
        short_usage = (
            clypi.style(option.short_display_name, fg="green", bold=True)
            if option.short
            else ""
        )
        type_str = clypi.style(
            type_util.type_to_str(option.arg_type).upper(), fg="yellow", bold=True
        )
        help = option.help or ""
        return usage, short_usage, type_str, help

    def _format_options(self) -> list[str] | None:
        if not self.options:
            return None

        usage: list[str] = []
        short_usage: list[str] = []
        type_str: list[str] = []
        help: list[str] = []
        for o in self.options:
            u, su, ts, hp = self._format_option(o)
            usage.append(u)
            short_usage.append(su)
            type_str.append(ts)
            help.append(hp)
        return list(boxed(stack(usage, short_usage, type_str, help), title="Options"))

    def _format_positional(self, positional: Argument) -> t.Any:
        name = clypi.style(positional.name, fg="blue", bold=True)
        help = positional.help or ""
        type_str = clypi.style(
            type_util.type_to_str(positional.arg_type).upper(), fg="yellow", bold=True
        )
        return name, type_str, help

    def _format_positionals(self) -> list[str] | str | None:
        if not self.positionals:
            return None

        name: list[str] = []
        type_str: list[str] = []
        help: list[str] = []
        for p in self.positionals:
            n, ts, hp = self._format_positional(p)
            name.append(n)
            type_str.append(ts)
            help.append(hp)
        return list(boxed(stack(name, type_str, help), title="Arguments"))

    def _format_subcommand(self, subcmd: SubCommand) -> t.Any:
        name = clypi.style(subcmd.name, fg="blue", bold=True)
        help = subcmd.help or ""
        return name, help

    def _format_subcommands(self) -> list[str] | str | None:
        if not self.subcommands:
            return None

        name: list[str] = []
        help: list[str] = []
        for p in self.subcommands:
            n, hp = self._format_subcommand(p)
            name.append(n)
            help.append(hp)
        return list(boxed(stack(name, help), title="Subcommands"))

    def _format_header(self) -> list[str] | str | None:
        prefix = clypi.style("Usage:", fg="yellow")
        prog = clypi.style(" ".join(self.prog), bold=True)

        options = (
            " [" + clypi.style("OPTIONS", fg="blue", bold=True) + "]"
            if self.options
            else ""
        )
        command = (
            clypi.style(" COMMAND", fg="blue", bold=True) if self.subcommands else ""
        )
        positional = (
            " "
            + " ".join(
                clypi.style(p.name.upper(), fg="blue", bold=True)
                for p in self.positionals
            )
            if self.positionals
            else ""
        )

        return [f"{prefix} {prog}{options}{command}{positional}"]

    def _format_description(self) -> list[str] | str | None:
        if not self.description:
            return None
        return [self.description, ""]

    def _format_exception(self) -> list[str] | str | None:
        if not self.exception:
            return ""
        return list(
            boxed(_pretty_traceback(self.exception), title="Error", color="red")
        )

    def format_help(self) -> str:
        lines: list[str] = []

        # Header
        _ext(lines, self._format_header())
        _ext(lines, "")

        # Description
        _ext(lines, self._format_description())

        # Options
        _ext(lines, self._format_options())

        # Positionals
        _ext(lines, self._format_positionals())

        # Subcommands
        _ext(lines, self._format_subcommands())

        # exceptions
        _ext(lines, self._format_exception())

        return "\n".join(lines)



from clypi import Command, config


class Lint(Command):
    files: tuple[str, ...]

    async def run(self, root):
        print(f"Linting {', '.join(self.files)}")


class MyCli(Command):
    subcommand: Lint | None = None
    verbose: bool = config(short="v", default=False)

    async def run(self, root):
        print(f"Running the main command with {self.verbose}")


if __name__ == "__main__":
    cli: MyCli = MyCli.parse()
    cli.start()



import asyncio
import sys
import typing as t
from contextlib import AbstractAsyncContextManager

from typing_extensions import override

import clypi
from clypi._data.spinners import Spin as _Spin
from clypi.colors import ColorType
from clypi.const import ESC

MOVE_START = f"{ESC}1G"
DEL_LINE = f"{ESC}0K"

Spin = _Spin


@t.final
class Spinner(AbstractAsyncContextManager):
    def __init__(
        self,
        title: str,
        animation: Spin | list[str] = Spin.DOTS,
        prefix: str = " ",
        suffix: str = "…",
        speed: float = 1,
    ) -> None:
        self.animation = animation
        self.prefix = prefix
        self.suffix = suffix
        self.title = title

        self._task: asyncio.Task[None] | None = None
        self._manual_exit: bool = False
        self._frame_idx: int = 0
        self._refresh_rate = 0.7 / speed / len(self._frames)

    async def __aenter__(self):
        self._task = asyncio.create_task(self._spin())
        return self

    def _print(
        self,
        msg: str,
        icon: str | None = None,
        color: ColorType | None = None,
        end: str = "",
    ):
        # Build the line being printed
        icon = clypi.style(icon + " ", fg=color) if icon else ""
        msg = f"{self.prefix}{icon}{msg}{end}"

        # Wipe the line for next render
        sys.stdout.write(MOVE_START)
        sys.stdout.write(DEL_LINE)

        # Write msg and flush
        sys.stdout.write(msg)
        sys.stdout.flush()

    def _render_frame(self):
        self._print(
            self.title + self.suffix,
            icon=self._frames[self._frame_idx],
            color="blue",
        )

    @property
    def _frames(self) -> list[str]:
        return (
            self.animation.value if isinstance(self.animation, Spin) else self.animation
        )

    async def _spin(self) -> None:
        while True:
            self._frame_idx = (self._frame_idx + 1) % len(self._frames)
            self._render_frame()
            await asyncio.sleep(self._refresh_rate)

    @override
    async def __aexit__(self, _type, value, traceback):
        # If a user already called `.done()`, leaving the closure
        # should not re-trigger a re-render
        if self._manual_exit:
            return

        if any([_type, value, traceback]):
            self.fail()
        else:
            self.done()

    def _exit(self, msg: str | None = None, success: bool = True):
        if t := self._task:
            t.cancel()

        color = "green" if success else "red"
        icon = "✔️" if success else "×"
        self._print(msg or self.title, icon=icon, color=color, end="\n")

    def done(self, msg: str | None = None):
        self._manual_exit = True
        self._exit(msg)

    def fail(self, msg: str | None = None):
        self._manual_exit = True
        self._exit(msg, success=False)

    def log(self, msg: str):
        self._print(msg.rstrip(), end="\n")
        self._render_frame()

    async def pipe(
        self,
        pipe: asyncio.StreamReader | None,
        color: ColorType = "blue",
        prefix: str = "",
    ) -> None:
        if not pipe:
            return

        while True:
            line = await pipe.readline()
            if not line:
                break

            icon = f"   ┃ {prefix}" if prefix else "   ┃"
            self._print(
                line.decode(),
                icon=icon,
                color=color,
            )
            self._render_frame()



import asyncio
from pathlib import Path
from unittest.mock import MagicMock, patch

from typing_extensions import override

from clypi import Command, config


class ExampleSubCommand(Command):
    """Some sample docs"""

    positional: tuple[str | Path, ...]

    async def run(self, root):
        return "subcommand"


class ExampleCommand(Command):
    """
    Some sample documentation for the main command
    """

    flag: bool = False
    subcommand: ExampleSubCommand | None = None
    option: list[str] = config(help="A list of strings please", default_factory=list)

    @override
    @classmethod
    def prog(cls):
        return "example"

    @override
    @classmethod
    def epilog(cls):
        return "Some text to display after..."

    async def run(self, root):
        return "main"


def test_expected_base():
    assert ExampleCommand.help() == "Some sample documentation for the main command"
    assert ExampleCommand.prog() == "example"
    assert ExampleCommand.epilog() == "Some text to display after..."


def test_expected_options():
    opts = ExampleCommand.options()
    assert len(opts) == 2

    assert opts["flag"].name == "flag"
    assert opts["flag"].arg_type is bool
    assert opts["flag"].nargs == 0

    assert opts["option"].name == "option"
    assert opts["option"].arg_type == list[str]
    assert opts["option"].nargs == "*"


def test_expected_positional():
    pos = ExampleSubCommand.positionals()
    assert len(pos) == 1

    assert pos["positional"].name == "positional"
    assert pos["positional"].arg_type == tuple[str | Path, ...]
    assert pos["positional"].nargs == "+"


def test_expected_subcommands():
    pos = ExampleCommand.subcommands()
    assert len(pos) == 1

    assert pos["example-sub-command"].name == "example-sub-command"
    assert pos["example-sub-command"].klass == ExampleSubCommand
    assert pos["example-sub-command"].help == "Some sample docs"


@patch("os.get_terminal_size")
def test_expected_parsing(gts):
    gts.return_value = MagicMock()
    gts.return_value.columns = 80

    ec = ExampleCommand.parse(["--flag", "--option", "a", "b"])
    assert ec.flag is True
    assert ec.option == ["a", "b"]

    assert ec.subcommand is None
    assert asyncio.run(ec.astart()) == "main"


@patch("os.get_terminal_size")
def test_expected_parsing_subcmd(gts):
    gts.return_value = MagicMock()
    gts.return_value.columns = 80

    ec = ExampleCommand.parse(
        ["--flag", "--option", "a", "b", "example-sub-command", "some_file.json"]
    )
    assert ec.flag is True
    assert ec.option == ["a", "b"]

    sc = ec.subcommand
    assert isinstance(sc, ExampleSubCommand)
    assert sc.positional == ("some_file.json",)

    assert asyncio.run(ec.astart()) == "subcommand"



from __future__ import annotations

from typing import Generator, cast

import clypi
from clypi.colors import ColorType, _color_codes


# --- DEMO UTILS ---
def _all_colors() -> Generator[tuple[ColorType, ...], None, None]:
    for color in _color_codes:
        yield (
            cast(ColorType, f"{color}"),
            cast(ColorType, f"bright_{color}"),
        )


# --- DEMO START ---
def main() -> None:
    fg_block = []
    for color, bright_color in _all_colors():
        fg_block.append(
            clypi.style("██ " + color.ljust(9), fg=color)
            + clypi.style("██ " + bright_color.ljust(16), fg=bright_color)
        )

    bg_block = []
    for color, bright_color in _all_colors():
        bg_block.append(
            clypi.style(color.ljust(9), bg=color)
            + " "
            + clypi.style(bright_color.ljust(16), bg=bright_color)
        )

    style_block = []
    style_block.append(clypi.style("I am bold", bold=True))
    style_block.append(clypi.style("I am dim", dim=True))
    style_block.append(clypi.style("I am underline", underline=True))
    style_block.append(clypi.style("I am blink", blink=True))
    style_block.append(clypi.style("I am reverse", reverse=True))
    style_block.append(clypi.style("I am strikethrough", strikethrough=True))

    for line in clypi.stack(
        clypi.boxed(fg_block, width=35, title="Foregrounds"),
        clypi.boxed(bg_block, width=30, title="Backgrounds"),
        clypi.boxed(style_block, width=22, title="Styles"),
    ):
        print(line)


if __name__ == "__main__":
    main()



from __future__ import annotations

import asyncio
import inspect
import logging
import re
import sys
import typing as t
from dataclasses import dataclass
from types import NoneType, UnionType

from Levenshtein import distance  # type: ignore

from clypi._cli import config as _conf
from clypi._cli import parser, type_util
from clypi._cli.formatter import TermFormatter

logger = logging.getLogger(__name__)

# re-exports
config = _conf.config


def _camel_to_dashed(s: str):
    return re.sub(r"(?<!^)(?=[A-Z])", "-", s).lower()


@dataclass
class Argument:
    name: str
    arg_type: t.Any
    help: str | None
    is_opt: bool = False
    short: str | None = None

    @property
    def nargs(self) -> parser.Nargs:
        if self.arg_type is bool:
            return 0

        if type_util.is_collection(self.arg_type):
            return "*"

        if type_util.is_tuple(self.arg_type):
            sz = type_util.tuple_size(self.arg_type)
            return "+" if sz == float("inf") else sz

        return 1

    @property
    def display_name(self):
        name = parser.snake_to_dash(self.name)
        if self.is_opt:
            return f"--{name}"
        return name

    @property
    def short_display_name(self):
        assert self.short
        name = parser.snake_to_dash(self.short)
        return f"-{name}"


@dataclass
class SubCommand:
    name: str
    klass: type[Command]
    help: str


class Command:
    @t.final
    @classmethod
    def name(cls):
        return _camel_to_dashed(cls.__name__)

    @classmethod
    def prog(cls) -> str:
        return cls.name()

    @classmethod
    def epilog(cls) -> str | None:
        return None

    @t.final
    @classmethod
    def help(cls):
        doc = inspect.getdoc(cls)

        # Dataclass sets a default docstring so ignore that
        if not doc or doc.startswith(cls.__name__ + "("):
            return None

        return doc.replace("\n", " ")

    async def run(self, root: Command) -> None:
        """
        This function is where the business logic of your command
        should live.

        `self` contains the arguments for this command you can access
        as any other instance property.

        `root` is a pointer to the base command of your CLI so that you
        can access arguments passed to parent commands.
        """
        raise NotImplementedError

    @t.final
    async def astart(self, root: Command | None = None) -> None:
        if subcommand := getattr(self, "subcommand", None):
            return await subcommand.astart(root=root or self)
        return await self.run(root or self)

    @t.final
    def start(self) -> None:
        asyncio.run(self.astart())

    @t.final
    @classmethod
    def fields(cls) -> dict[str, _conf.Config[t.Any]]:
        """
        Parses the type hints from the class extending Command and assigns each
        a _Config field with all the necessary info to display and parse them.
        """
        defaults: dict[str, _conf.Config[t.Any]] = {}
        for field, _type in inspect.get_annotations(cls).items():
            default = getattr(cls, field, _conf.MISSING)
            if isinstance(default, _conf.PartialConfig):
                defaults[field] = _conf.Config.from_partial(
                    default,
                    parser=default.parser or parser.from_type(_type),
                    arg_type=_type,
                )
            else:
                defaults[field] = _conf.Config(
                    default=default,
                    parser=parser.from_type(_type),
                    arg_type=_type,
                )
        return defaults

    @t.final
    @classmethod
    def _next_positional(cls, kwargs: dict[str, t.Any]) -> Argument | None:
        """
        Traverse the current collected arguments and find the next positional
        arg we can assign to.
        """
        for pos in cls.positionals().values():
            # List positionals are a catch-all
            if type_util.is_collection(pos.arg_type):
                return pos

            if pos.name not in kwargs:
                return pos

    @t.final
    @classmethod
    def _get_long_name(cls, short: str) -> str | None:
        fields = cls.fields()
        for field, field_conf in fields.items():
            if field_conf.short == short:
                return field
        return None

    @t.final
    @classmethod
    def _find_similar_arg(cls, arg: parser.Arg) -> str | None:
        """
        Utility function to find arguments similar to the one the
        user passed in to correct typos.
        """
        if arg.is_pos():
            for pos in cls.subcommands().values():
                if distance(pos.name, arg.value) < 3:
                    return pos.name

            for pos in cls.positionals().values():
                if distance(pos.name, arg.value) < 3:
                    return pos.display_name
        else:
            for opt in cls.options().values():
                if distance(opt.name, arg.value) <= 2:
                    return opt.display_name
                if opt.short and distance(opt.short, arg.value) <= 1:
                    return opt.short_display_name

        return None

    @t.final
    @classmethod
    def _safe_parse(cls, args: t.Iterator[str], parents: list[str]) -> t.Self:
        """
        Tries parsing args and if an error is shown, it displays the subcommand
        that failed the parsing's help page.
        """
        try:
            return cls._parse(args, parents)
        except (ValueError, TypeError) as e:
            cls.print_help(parents, exception=e)

    @t.final
    @classmethod
    def _parse(cls, args: t.Iterator[str], parents: list[str]) -> t.Self:
        """
        Given an iterator of arguments we recursively parse all options, arguments,
        and subcommands until the iterator is complete.
        """

        # The kwars used to initialize the dataclass
        kwargs: dict[str, t.Any] = {}

        # The current option or positional arg being parsed
        current_attr = parser.CurrentCtx()

        def flush_ctx():
            nonlocal current_attr
            if current_attr and current_attr.needs_more():
                raise ValueError(f"Not enough values for {current_attr.name}")
            elif current_attr:
                kwargs[current_attr.name] = current_attr.collected
                current_attr = None

        def find_similar(parsed: parser.Arg):
            what = "argument" if parsed.is_pos() else "option"
            error = f"Unknown {what} {parsed.orig!r}"
            if similar := cls._find_similar_arg(parsed):
                error += f". Did you mean {similar!r}?"
            raise ValueError(error)

        for a in args:
            if a in ("-h", "--help"):
                cls.print_help(parents=parents)

            # ---- Try to parse as an arg/opt ----
            parsed = parser.parse_as_attr(a)
            if parsed.is_pos() and (subcmd := cls.subcommands().get(parsed.value)):
                kwargs["subcommand"] = subcmd.klass._safe_parse(
                    args, parents=parents + [cls.prog()]
                )
                break

            # ---- Try to set to the current option ----
            is_valid_long = (
                not parsed.is_pos()
                and parsed.is_long_opt()
                and parsed.value in cls.options()
            )
            is_valid_short = (
                not parsed.is_pos()
                and parsed.is_short_opt()
                and cls._get_long_name(parsed.value) is not None
            )
            if (
                parsed.is_short_opt()
                or parsed.is_long_opt()
                and not (is_valid_long or is_valid_short)
            ):
                raise find_similar(parsed)

            if is_valid_long or is_valid_short:
                long_name = cls._get_long_name(parsed.value) or parsed.value
                option = cls.options()[long_name]
                flush_ctx()

                # Boolean flags don't need to parse more args later on
                if option.nargs == 0:
                    kwargs[long_name] = True
                else:
                    current_attr = parser.CurrentCtx(
                        option.name, option.nargs, option.nargs
                    )
                continue

            # ---- Try to assign to the current positional ----
            if not current_attr.name and (pos := cls._next_positional(kwargs)):
                current_attr = parser.CurrentCtx(pos.name, pos.nargs, pos.nargs)

            # ---- Try to assign to the current ctx ----
            if current_attr.name and current_attr.has_more():
                current_attr.collect(parsed.value)
                continue

            raise find_similar(parsed)

        # If we finished the loop but an option needs more args, fail
        if current_attr.name and current_attr.needs_more():
            raise ValueError(f"Not enough values for {current_attr.name}")

        # If we finished the loop and we haven't saved current_attr, save it
        if current_attr.name and not current_attr.needs_more():
            kwargs[current_attr.name] = current_attr.collected
            current_attr = None

        # Parse as the correct values and assign to the instance
        instance = cls()
        for field, field_conf in cls.fields().items():
            if field not in kwargs and not field_conf.has_default():
                raise ValueError(f"Missing required argument {field}")

            # Get the value passed in or the provided default
            value = kwargs[field] if field in kwargs else field_conf.get_default()

            # Subcommands are already parsed properly
            if field == "subcommand":
                setattr(instance, field, value)
                continue

            # Try parsing the string as the right type
            parsed = field_conf.parser(value)
            setattr(instance, field, parsed)

        return instance

    @t.final
    @classmethod
    def subcommands(cls) -> dict[str, SubCommand]:
        if "subcommand" not in cls.fields():
            return {}

        # Get the subcommand type/types
        _type = cls.fields()["subcommand"].arg_type
        subcmds = [_type]
        if isinstance(_type, UnionType):
            subcmds = [s for s in _type.__args__ if s is not NoneType]

        for v in subcmds:
            assert inspect.isclass(v) and issubclass(v, Command)

        return {
            s.name(): SubCommand(name=s.name(), klass=s, help=s.help()) for s in subcmds
        }

    @t.final
    @classmethod
    def options(cls) -> dict[str, Argument]:
        options: dict[str, Argument] = {}
        for field, field_conf in cls.fields().items():
            if field == "subcommand" or not field_conf.has_default():
                continue

            options[field] = Argument(
                field,
                type_util.remove_optionality(field_conf.arg_type),
                help=field_conf.help,
                short=field_conf.short,
                is_opt=True,
            )
        return options

    @t.final
    @classmethod
    def positionals(cls) -> dict[str, Argument]:
        options: dict[str, Argument] = {}
        for field, field_conf in cls.fields().items():
            if field == "subcommand" or field_conf.has_default():
                continue

            options[field] = Argument(
                field,
                field_conf.arg_type,
                help=field_conf.help,
            )
        return options

    @t.final
    @classmethod
    def parse(cls, args: t.Sequence[str] | None = None) -> t.Self:
        """
        This is the entry point to start parsing arguments
        """
        norm_args = parser.normalize_args(args or sys.argv[1:])
        args_iter = iter(norm_args)
        instance = cls._safe_parse(args_iter, parents=[])
        if list(args_iter):
            raise ValueError(f"Unknown arguments {list(args_iter)}")

        return instance

    @t.final
    @classmethod
    def print_help(cls, parents: list[str] = [], *, exception: Exception | None = None):
        tf = TermFormatter(
            prog=parents + [cls.prog()],
            description=cls.help(),
            epilog=cls.epilog(),
            options=list(cls.options().values()),
            positionals=list(cls.positionals().values()),
            subcommands=list(cls.subcommands().values()),
            exception=exception,
        )
        print(tf.format_help())
        sys.exit(1 if exception else 0)

    def __repr__(self) -> str:
        fields = ", ".join(
            f"{k}={v}"
            for k, v in vars(self).items()
            if v is not None and not k.startswith("_")
        )
        return f"{self.__class__.__name__}({fields})"



import typing as t
from dataclasses import asdict, dataclass

T = t.TypeVar("T")


class _MISSING_TYPE:
    pass


MISSING = _MISSING_TYPE()


@dataclass
class PartialConfig(t.Generic[T]):
    parser: t.Callable[[t.Any], T] | None = None
    default: T | _MISSING_TYPE = MISSING
    default_factory: t.Callable[[], T] | _MISSING_TYPE = MISSING
    help: str | None = None
    short: str | None = None

    def has_default(self) -> bool:
        return self.default is not MISSING or self.default_factory is not MISSING

    def get_default(self) -> T:
        if not isinstance(self.default, _MISSING_TYPE):
            return self.default

        if t.TYPE_CHECKING:
            assert not isinstance(self.default_factory, _MISSING_TYPE)

        return self.default_factory()


@dataclass
class Config(t.Generic[T]):
    parser: t.Callable[[t.Any], T]
    arg_type: t.Any
    default: T | _MISSING_TYPE = MISSING
    default_factory: t.Callable[[], T] | _MISSING_TYPE = MISSING
    help: str | None = None
    short: str | None = None

    def has_default(self) -> bool:
        return not isinstance(self.default, _MISSING_TYPE) or not isinstance(
            self.default_factory, _MISSING_TYPE
        )

    def get_default(self) -> T:
        if not isinstance(self.default, _MISSING_TYPE):
            return self.default

        if t.TYPE_CHECKING:
            assert not isinstance(self.default_factory, _MISSING_TYPE)

        return self.default_factory()

    @classmethod
    def from_partial(
        cls, partial: PartialConfig[T], parser: t.Callable[[t.Any], T], arg_type: t.Any
    ):
        kwargs = asdict(partial)
        kwargs.update(parser=parser, arg_type=arg_type)
        return cls(**kwargs)


def config(
    parser: t.Callable[[t.Any], T] | None = None,
    default: T | _MISSING_TYPE = MISSING,
    default_factory: t.Callable[[], T] | _MISSING_TYPE = MISSING,
    help: str | None = None,
    short: str | None = None,
) -> T:
    return PartialConfig(
        parser=parser,
        default=default,
        default_factory=default_factory,
        help=help,
        short=short,
    )  # type: ignore



from clypi.colors import remove_style


def _safe_get(ls: list[str], idx: int) -> str:
    if idx >= len(ls):
        return ""
    return ls[idx]


def _real_len(s: str) -> int:
    s = remove_style(s)
    return len(s)


def stack(*blocks: list[str], padding: int = 1) -> list[str]:
    new_lines = []
    height = max(len(b) for b in blocks)
    widths = [max(_real_len(line) for line in block) for block in blocks]

    # Process line until all blocks are done
    for idx in range(height):
        more = False
        tmp: list[str] = []

        # Add the line from each block
        for block, width in zip(blocks, widths):
            # If there was a line, next iter will happen
            block_line = _safe_get(block, idx)
            if block_line:
                more |= True

            # How much do we need to reach the actual visible length
            actual_width = (width - _real_len(block_line)) + len(block_line)

            # Align and append line
            tmp.append(block_line.ljust(actual_width))
            tmp.append(" " * padding)

        new_lines.append(" ".join(tmp))

        # Exit if no more lines in any iter
        if not more:
            break

    return new_lines



from dataclasses import dataclass
from enum import Enum


@dataclass
class Box:
    """
    tl  x myt x  tr
    y             y
    mxl x mm  x mxr
    y             y
    bl  x myb x  br
    """

    tl: str
    tr: str
    bl: str
    br: str
    x: str
    y: str
    myt: str
    myb: str
    mxl: str
    mxr: str
    mm: str


_HEAVY = Box(
    tl="┏",
    tr="┓",
    bl="┗",
    br="┛",
    x="━",
    y="┃",
    myt="┳",
    myb="┻",
    mxl="┣",
    mxr="┫",
    mm="╋",
)


class Boxes(Enum):
    HEAVY = _HEAVY



from __future__ import annotations

import io
import re
import sys
from contextlib import contextmanager
from datetime import date, datetime, timedelta

import pytest
from pytest import mark

import clypi
from clypi.prompts import MaxAttemptsException, Parser


@contextmanager
def replace_stdin(answers: list[str] | str = []):
    answers = [answers] if isinstance(answers, str) else answers
    target = "\n".join(answers)

    orig = sys.stdin
    sys.stdin = io.StringIO(target)
    yield
    sys.stdin = orig


@contextmanager
def replace_stdout():
    orig = sys.stdout
    new = io.StringIO()
    sys.stdout = new
    yield new
    sys.stdout = orig


def _escape_ansi(line: str) -> str:
    """
    https://stackoverflow.com/questions/14693701/how-can-i-remove-the-ansi-escape-sequences-from-a-string-in-python
    """
    ansi_escape = re.compile(r"(?:\x1B[@-_]|[\x80-\x9F])[0-?]*[ -/]*[@-~]")
    return ansi_escape.sub("", line)


def assert_prompted_times(prompted: io.StringIO, times: int):
    __tracebackhide__ = True
    text = _escape_ansi(prompted.getvalue())
    lines = list(filter(None, text.split(": ")))
    assert len(lines) == times


@mark.parametrize(
    "answer,expected",
    [
        ("Alice\n", "Alice"),
        ("\n", "John Doe"),
    ],
)
def test_prompt_with_default(answer: str, expected: str):
    with replace_stdin(answer) as _:
        assert clypi.prompt("What's your name?", default="John Doe") == expected


@mark.parametrize(
    "answers,expected,times",
    [
        (["Alice"], "Alice", 1),
        (["", "", "Alice"], "Alice", 3),
    ],
)
def test_prompt_with_no_default(answers: list[str], expected: str, times: int):
    with replace_stdin(answers) as _, replace_stdout() as stdout:
        assert clypi.prompt("What's your name?") == expected
        assert_prompted_times(stdout, times)


def test_prompt_with_parser():
    with replace_stdin("42") as _:
        res = clypi.prompt("Some prompt", parser=int)
        assert isinstance(res, int)


@mark.parametrize(
    "answer,parser",
    [
        ("Alice", int),
        ("42 days", int),
        ("42 asd", timedelta),
        ("42", timedelta),
        ("202-01-01", date),
        ("2021-01-01T00:00", datetime),
    ],
    ids=[
        "Str as Int",
        "TimeDelta as Int",
        "Invalid TimeDelta",
        "Int as TimeDelta",
        "Invalid Date",
        "Invalid DateTime",
    ],
)
def test_prompt_with_parser_fails(answer: str, parser: Parser):
    with replace_stdin(answer) as _, pytest.raises(MaxAttemptsException):
        clypi.prompt("Some prompt", parser=parser, max_attempts=1)


def test_prompt_with_good_parser():
    with replace_stdin("2") as _:
        res = clypi.prompt("Some prompt", parser=lambda x: int(x) * 2)
        assert res == 4


def _raise_error(x: int) -> None:
    raise ValueError(f"Invalid number {x}")


def test_prompt_with_bad_validate():
    with replace_stdin("2") as _, pytest.raises(MaxAttemptsException):
        clypi.prompt("Some prompt", parser=_raise_error, max_attempts=1)


def test_prompt_with_provided():
    with replace_stdin() as _:
        res = clypi.prompt("Some prompt", parser=int, provided=42)
        assert res == 42
        assert isinstance(res, int)
